<Window
       xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
       xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
       Title="PyAEDT WPF Application" Height="800" Width="600" Background="#FFD1CFCF">
       <Grid>
        <Image x:Name="logo" HorizontalAlignment="Right" Height="36" VerticalAlignment="Top" Width="99" Source="pyansys-logo-black-cropped.png" Margin="96,0,0,0"/>
    </Grid>
</Window>
